/* symo.cpp
 *
 * Sort Your MP3s Out, main database (UI-independent)
 * (K) All Rites Reversed - Copy What You Like (see file Copying)
 *
 * Authors:
 *      Peter Hartley       <pdh@chaos.org.uk>
 */

#include <ctype.h>
#include <stdio.h>
#include <sys/stat.h>
#include <dirent.h>
#include <algorithm>
#include <errno.h>
#include <unistd.h>
#include <sys/signal.h>
#include <sys/wait.h>

#include <id3/tag.h>
#include <id3/misc_support.h>

#include "symo.h"
#include "id3genres.h"

#include "empeg/mp3_frame_info.h"
#include "empeg/file_byte_stream.h"
#include "empeg/extract_tags.h"

static bool isSuspicious( const char *title )
{
    if ( islower( title[0] ) )
	return true;

    if ( strlen( title ) == 30 )
	return true; // ID3v1!

    int opens = 0, closes = 0;

    while ( *title )
    {
	if ( *title == ' ' && islower( title[1] ) )
	    return true;
	if ( *title == ' ' && (title[1] == '.' || title[1] == '\0') )
	    return true;
	if ( *title == '\'' && title[1] == '\'' )
	    return true;
	if ( isdigit(*title) && (title[1] == '\'' || title[1] == '\"' ) )
	    return true;
	if ( *title == '(' )
	    opens++;
	if ( *title == ')' )
	    closes++;
	title++;
    }
    return opens != closes;
}

extern "C" void sigchild_handler(int)
{
    pid_t pid;
    int stat;
    
    while ( (pid = waitpid(-1, &stat, WNOHANG)) > 0 )
    {
	printf("pid %d exited\n", pid);
    }
}

Symo::Symo()
    : m_editedCount(0)
{
    signal(SIGCHLD, &sigchild_handler);
}

Symo::~Symo()
{
}

class SymoTagsObserver: public ExtractTagsObserver
{
    Symo::Entry *m_entry;
public:
    explicit SymoTagsObserver(Symo::Entry* e) : m_entry(e) {}
    void OnExtractTag(const char *name, const char *value)
	{
	    if (!strcmp(name, "bitrate"))
		m_entry->bitrate = value;
	}
    void OnExtractTag(const char *name, int value)
	{
	    if (!strcmp(name, "duration"))
	    {
		// It's in ms
		m_entry->sec = value/1000;
	    }
	}
};

void Symo::addFile( const char *fname )
{
//    printf( "symo::addFile(%s)\n", fname );

    struct stat st;

    if ( stat( fname, &st ) < 0 )
	return;

    if ( S_ISDIR( st.st_mode ) )
    {
	DIR *dp;
	struct dirent *ep;

	dp = opendir( fname );
	if ( dp )
	{
	    while ( (ep = readdir(dp)) != NULL )
	    {
		if ( ep->d_name[0] != '.' )
		{
		    char buf[4096];
		    sprintf( buf, "%s/%s", fname, ep->d_name );
		    addFile( buf );
		}
	    }
	    closedir( dp );
	}
    }
    else if ( S_ISREG( st.st_mode ) )
    {
	Entry e;

	e.originalFilename = fname;
	e.mtime = st.st_mtime;
	m_vec.push_back( e );
    }
}

const char *safe( const char *s )
{
    return s ? s : "";
}

static bool EntryDeletedP(const Symo::Entry& e)
{
    return e.deleted;
}

void Symo::read( SymoObserver *o )
{
    size_t n = m_vec.size();
    size_t current = 0;

    if ( o )
	o->onActivity( 0, n );

    for ( vec_t::iterator i = m_vec.begin();
	  i != m_vec.end();
	  ++i )
    {
	Entry& e = *i;

	e.edited = false;
	e.deleted = false;
	e.filename = e.originalFilename;

	struct stat st;
	stat( e.filename.c_str(), &st );

//	printf( "id3ing %s\n", e.filename.c_str() );

	e.sz = st.st_size;
	e.sec = 0;
	e.bitrate = "";

	ID3_Tag tag( e.filename.c_str() );
	    
	e.artist = safe( ID3_GetArtist(&tag) );
	e.album  = safe( ID3_GetAlbum(&tag)  );
	e.title  = safe( ID3_GetTitle(&tag)  );
	e.year   = safe( ID3_GetYear(&tag)   );
	e.genre  = safe( ID3_GetGenre(&tag)  );
	e.trackno = safe( ID3_GetTrack(&tag) );
       
	e.artistStatus = PLAIN;
	e.albumStatus = PLAIN;
	e.titleStatus = PLAIN;
	e.yearStatus = PLAIN;
	e.genreStatus = SUSPICIOUS; // Until proven plain
	e.filenameStatus = PLAIN;
	
	if ( isSuspicious( e.filename.c_str() ) )
	    e.filenameStatus = SUSPICIOUS;
	if ( isSuspicious(e.artist.c_str()) )
	    e.artistStatus = SUSPICIOUS;
	else
	{
	    if ( e.filename.find( e.artist ) == string::npos )
	    {
		if ( !strncmp(e.artist.c_str(), "The ", 4)
		     && e.filename.find(e.artist.substr(4)) != string::npos )
		{
		}
		else
		    e.artistStatus = SUSPICIOUS;
	    }
	}
	if ( e.filename.find( e.album ) == string::npos
	     || isSuspicious(e.album.c_str()) )
	    e.albumStatus = SUSPICIOUS;
	if ( e.filename.find( e.title ) == string::npos
	     || e.title.find( ".mp3" ) != string::npos
	     || isSuspicious(e.title.c_str()) )
	    e.titleStatus = SUSPICIOUS;
	if ( e.year.length() != 4
	     || e.year.find_first_not_of("0123456789") != string::npos )
	    e.yearStatus = SUSPICIOUS;
	
	int genre_ix;
	if (sscanf(e.genre.c_str(), "(%d)", &genre_ix) == 1)
	{
	    e.genre = m_genres.FromNumber(genre_ix);
	    e.genreStatus = PLAIN;
	}
	else if (m_genres.FromString(e.genre) != -1)
	    e.genreStatus = PLAIN;

//	printf( "Genre '%s' status %d\n", e.genre.c_str(),
//		e.genreStatus );

	SymoTagsObserver sto(&e);
	ReadFileStream fs;
	if (SUCCEEDED(fs.Open(e.originalFilename.c_str())))
	{
	    FrameInfoExtract fie(&fs);
	    if (FAILED(fie.Extract(0,&sto)) || e.bitrate == "")
	    {
		printf( "%s probably not MP3\n", e.originalFilename.c_str());
		e.deleted = true;
	    }
	}

	current++;
	if ( o )
	    o->onActivity( current, n );
    }

    m_vec.erase(std::remove_if(m_vec.begin(), m_vec.end(), EntryDeletedP), m_vec.end());
}

void Symo::changeEntry( unsigned int i, const Entry& new_e )
{
    if ( i>=entries() )
	return;

    Entry& e = m_vec[i];

    bool any = false;

    if ( new_e.filename != e.filename )
    {
	e.filename = new_e.filename;
	e.filenameStatus = EDITED;
	any = true;
    }
    if ( new_e.title != e.title )
    {
	e.title = new_e.title;
	e.titleStatus = EDITED;
	any = true;
    }
    if ( new_e.artist != e.artist )
    {
	e.artist = new_e.artist;
	e.artistStatus = EDITED;
	any = true;
    }
    if ( new_e.album != e.album )
    {
	e.album = new_e.album;
	e.albumStatus = EDITED;
	any = true;
    }
    if ( new_e.year != e.year )
    {
	e.year = new_e.year;
	e.yearStatus = EDITED;
	any = true;
    }
    if ( new_e.genre != e.genre )
    {
	e.genre = new_e.genre;
	e.genreStatus = EDITED;
	any = true;
    }

    if ( new_e.trackno != e.trackno )
    {
	e.trackno = new_e.trackno;
	e.tracknoStatus = EDITED;
	any = true;
    }

    if ( any )
    {
	if ( !e.edited )
	    m_editedCount++;
	e.edited = true;
    }
}

static std::string upperCase( const std::string& s )
{
    std::string result(s);

    for ( unsigned int i=0; i<result.size(); ++i )
    {
	if ( isupper( result[i] ) )
	    result[i] = tolower( result[i] );
    }
    return result;
}

static std::string titleCase( const std::string& s )
{
    std::string result(s);
    bool new_word = true;

    for ( unsigned int i=0; i<result.size(); ++i )
    {
	if ( new_word )
	{
	    if ( islower( result[i] ) )
		result[i] = toupper( result[i] );
	}
/*	else
	{
	    if ( isupper( result[i] ) )
		result[i] = tolower( result[i] );
		} */
	new_word = isspace( result[i] );
    }
    return result;
}

static std::string findFileStub( const Symo::Entry& e )
{
    std::string::size_type pos, pos2 = e.filename.size();

    std::string upper = upperCase( e.filename );
    
    pos = upper.find( upperCase( e.artist ) );
    if ( pos < pos2 )
	pos2 = pos;

    pos = upper.find( upperCase( e.title ) );
    if ( pos < pos2 )
	pos2 = pos;

    pos = upper.find( upperCase( e.album ) );
    if ( pos < pos2 )
	pos2 = pos;

    pos = e.filename.rfind( '/', pos2 );
    if ( pos == std::string::npos )
	return "";

    std::string result( e.filename, 0, pos+1 );

//    printf( "STUB(%s) = '%s'\n", e.filename.c_str(), result.c_str() );

    return result;
}

void Symo::autoFixup( unsigned int which )
{
    Entry e = m_vec[which]; // copy constructor

    std::string filestub = findFileStub( e );

    e.title = titleCase( e.title );
    e.artist = titleCase( e.artist );
    e.album = titleCase( e.album );

    std::string artist_no_the = e.artist;
    if (!strncmp(artist_no_the.c_str(),"The ",4))
	artist_no_the = e.artist.substr(4);

    std::string filename = filestub + artist_no_the + "/";
    if ( e.album != "" )
    {
	filename += e.album + "/";
    }
    if ( e.trackno != "" )
    {
	char buf[16];
	unsigned int trackno = atoi(e.trackno.c_str());
	sprintf( buf, "%02d ", trackno );
	filename += buf;
    }
    filename += e.title + ".mp3";

    e.filename = filename;

    changeEntry( which, e );
}

void Symo::play( unsigned int which )
{
    int rc = fork();
    if (rc<0)
    {
	printf("Fork failed\n");
	return;
    }
    if (rc == 0)
    {
	/* Child */
	printf("Playing %s via freeamp\n", 
	       m_vec[which].originalFilename.c_str() );
	for (int i=3; i<64; i++)
	    close(i);

	rc = execlp("freeamp", "freeamp",
		    m_vec[which].originalFilename.c_str(), NULL);

	printf("Exec of freamp failed (%d)\n", errno);
	exit(0);
    }
    /* Parent -- do nothing */
}

static int ensureDirectory( const char* d )
{
    struct stat st;

    if ( stat( d, &st ) == 0
	 && S_ISDIR( st.st_mode ) )
	return 0;

    return mkdir( d, 0755 );
}

static std::string cleverRename( std::string from, std::string to )
{
    string::size_type pos = 0;

    while ( (pos = to.find('/',pos+1)) != string::npos )
    {
	string dirname = to.substr( 0, pos );
	int rc = ensureDirectory( dirname.c_str() );
	if ( rc < 0 )
	    return string("Can't make directory ") + dirname + ": "
		+ strerror(errno);
    }

    int rc = rename( from.c_str(), to.c_str() );
    if ( rc < 0 )
	return string("Can't rename ") + from + ": "
	    + strerror(errno);

    return "";
}

void Symo::write( SymoObserver *o )
{
    size_t n = entries();
    size_t nEdited = m_editedCount;
    size_t progress = 0;

    if ( o )
	o->onActivity( 0, nEdited );

    for ( unsigned int i=0; i<n; i++ )
    {
	Entry& e = m_vec[i];

	if ( e.edited )
	{
	    ID3_Tag tag;

	    tag.Link( e.originalFilename.c_str() );
	    tag.SetPadding( false );

	    bool doit = false;
	    bool ok = true;

	    if ( e.titleStatus == EDITED )
	    {
		if ( e.title.length() )
		    ID3_AddTitle( &tag, e.title.c_str(), true );
		else
		    ID3_RemoveTitles( &tag );
		doit = true;
	    }
	    if ( e.artistStatus == EDITED )
	    {
		if ( e.artist.length() )
		    ID3_AddArtist( &tag, e.artist.c_str(), true );
		else
		    ID3_RemoveArtists( &tag );
		doit = true;
	    }
	    if ( e.albumStatus == EDITED )
	    {
		if ( e.album.length() )
		{
		    //printf( "setting album to '%s'\n", e.album.c_str() );
		    ID3_AddAlbum( &tag, e.album.c_str(), true );
		}
		else
		{
		    //int x = 
		    ID3_RemoveAlbums( &tag );
		    //printf( "removeAlbums returned %d\n", x );
		    
		    //const char *y = ID3_GetAlbum( &tag );
		    //printf( "getAlbum returned %s\n", y ? y : "NULL" );
		}
		doit = true;
	    }
	    if ( e.yearStatus == EDITED )
	    {
		if ( e.year.length() )
		    ID3_AddYear( &tag, e.year.c_str(), true );
		else
		    ID3_RemoveYears( &tag );
		doit = true;
	    }
	    if ( e.genreStatus == EDITED )
	    {
		if ( e.genre.length() )
		    ID3_AddGenre( &tag, m_genres.FromString(e.genre), true );
		else
		    ID3_RemoveGenres( &tag );
		doit = true;
	    }
	    if ( e.tracknoStatus == EDITED )
	    {
		if ( e.trackno.length() )
		    ID3_AddTrack( &tag, atoi(e.trackno.c_str()), 0, true );
		else
		    ID3_RemoveTracks( &tag );
		doit = true;
	    }

	    if ( doit )
	    {
		// General tidying up
		ID3_AddArtist( &tag, ID3_GetArtist(&tag), true );
		ID3_AddAlbum( &tag, ID3_GetAlbum(&tag), true );
		ID3_AddTitle( &tag, ID3_GetTitle(&tag), true );
		ID3_AddYear( &tag, ID3_GetYear(&tag), true );
		//ID3_AddGenre( &tag, ID3_GetGenre(&tag), true );

		// Do this twice to avoid bug in id3lib v3.7.13 et pre
		ID3_RemoveComments( &tag, NULL );
		ID3_RemoveComments( &tag, NULL );

		try {
		    // Clear nasty Lyrics tags which id3lib doesn't deal with
		    if ( tag.HasTagType(ID3TT_LYRICS3)
			 || tag.HasTagType(ID3TT_MUSICMATCH)
			 || tag.HasTagType(ID3TT_LYRICS3V2) )
			luint bar = tag.Strip(ID3TT_APPENDED);
		    //printf( "tag.Strip returned %lu\n", bar );
		    luint foo = tag.Update(ID3TT_ID3);
		    //printf( "tag.Update() returned %lu\n", foo );
		    e.titleStatus = e.artistStatus =
			e.albumStatus = e.yearStatus = 
			e.genreStatus = PLAIN;

		    // id3lib likes removing permissions :-(
		    chmod( e.originalFilename.c_str(), 0644 );
		}
		catch ( ID3_Error err )
		{
		    cout << err.GetErrorFile() << " (" << err.GetErrorLine()
			 << "): "
			 << err.GetErrorType() << ": " << err.GetErrorDesc()
			 << endl;
		    
		    if ( o )
			o->onError( e.originalFilename + ": "
				    + err.GetErrorType() + ": "
				    + err.GetErrorDesc() );

		    ok = false;
		}
	    }

	    if ( e.filenameStatus == EDITED )
	    {
		string err = cleverRename( e.originalFilename, e.filename );
		if ( err == "" )
		{
		    e.filenameStatus = PLAIN;
		    e.originalFilename = e.filename;
		}
		else
		{
		    if ( o )
			o->onError( err );
		    ok = false;
		}
	    }

	    if ( ok )
	    {
		e.edited = false;
		m_editedCount--;
	    }

	    progress++;
	    if ( o )
	    {
		o->onEntryChanged( i );
		o->onActivity( progress, nEdited );
	    }
	}
    }
}

class SymoCompareString
{
    std::string Symo::Entry::*m_f;
    bool m_ascending;
public:
    SymoCompareString( std::string Symo::Entry::*f, bool ascending )
	: m_f(f), m_ascending(ascending) {}
    
    bool operator()( const Symo::Entry& x, const Symo::Entry& y )
	{
	    if ( m_ascending )
		return x.*m_f < y.*m_f;
	    else
		return y.*m_f < x.*m_f;
	}
};

class SymoCompareInt
{
    size_t Symo::Entry::*m_f;
    bool m_ascending;
public:
    SymoCompareInt( size_t Symo::Entry::*f, bool ascending )
	: m_f(f), m_ascending(ascending) {}
    
    bool operator()( const Symo::Entry& x, const Symo::Entry& y )
	{
	    if ( m_ascending )
		return x.*m_f < y.*m_f;
	    else
		return y.*m_f < x.*m_f;
	}
};

void Symo::sortEntries( Field f, bool ascending )
{
    if ( f == MMSS || f == SIZE )
    {
	size_t Entry::*ptr;
	switch ( f )
	{
	case MMSS: ptr = &Entry::sec; break;
	case SIZE: ptr = &Entry::sz; break;
	default:   return;
	}

	SymoCompareInt sc(ptr,ascending);
	::stable_sort( m_vec.begin(), m_vec.end(), sc );
    }
    else
    {
	std::string Entry::*ptr;

	switch ( f )
	{
	case FILENAME: ptr = &Entry::filename; break;
	case ARTIST:   ptr = &Entry::artist; break;
	case TITLE:    ptr = &Entry::title; break;
	case ALBUM:    ptr = &Entry::album; break;
	case YEAR:     ptr = &Entry::year; break;
	default:       return;
	}

	SymoCompareString sc(ptr,ascending);

	::stable_sort( m_vec.begin(), m_vec.end(), sc );
    }
}

/* eof */
