/* symo.cpp
 *
 * Sort Your MP3s Out, main program
 * (K) All Rites Reversed - Copy What You Like (see file Copying)
 *
 * Authors:
 *      Peter Hartley       <pdh@chaos.org.uk>
 */

#include <ctype.h>
#include <stdio.h>

#include <qfontmetrics.h>
#include <qapplication.h>
#include <qwindowsstyle.h>
#include <qprogressdialog.h>
#include <qmessagebox.h>
#include <qlayout.h>
#include <qtextview.h>
#include <qpopupmenu.h>

#include "symo.h"
#include "symo_qt.h"
#include "version.h"

static const char *const mini_exclam_xpm[] = {
"16 14 5 1",
" 	c None",
".	c gray50",
"X	c #008000",
"o	c white",
"O	c black",
"        .X      ",
"       .oXX.    ",
"      .oXXXO    ",
"      XXXXXO    ",
"     .XXXXO.    ",
"     XXXXXO     ",
"     XXXXO      ",
"     XXXO       ",
"     XXO        ",
"     XO         ",
"   .oXX         ",
"   XXXX.        ",
"    XXO.        ",
"    .O.         "};

SymoGrid::SymoGrid( Symo *symo, QWidget *parent )
    : QTable(0,NUM_COLUMNS,parent), m_nFiles(0),
      m_pmEdited( (const char**)mini_exclam_xpm ), m_symo(symo),
      m_lastSortCol(-1)
{
    horizontalHeader()->setLabel(COL_FILENAME,"Filename");
    horizontalHeader()->setLabel(COL_EDITED,"!");
    horizontalHeader()->setLabel(COL_SIZE,"Size");
    horizontalHeader()->setLabel(COL_MMSS,"mm:ss");
    horizontalHeader()->setLabel(COL_BITRATE,"Bitrate");
    horizontalHeader()->setLabel(COL_ARTIST,"Artist");
    horizontalHeader()->setLabel(COL_TRACK,"Track#");
    horizontalHeader()->setLabel(COL_ALBUM,"Album");
    horizontalHeader()->setLabel(COL_TITLE,"Title");
    horizontalHeader()->setLabel(COL_YEAR,"Year");
    horizontalHeader()->setLabel(COL_GENRE,"Genre");
    horizontalHeader()->setLabel(COL_MTIME,"Modified");
    setSorting(true);
    hideColumn(COL_INDEX);
    setLeftMargin(0);
    verticalHeader()->hide();

    connect( this, SIGNAL(valueChanged(int,int)),
	     this, SLOT(onValueChanged(int,int)) );
}

SymoGrid::~SymoGrid()
{
}

void SymoGrid::addEntry( const Symo::Entry&e )
{
    setNumRows( numRows() + 1 );

    char szbuf[16];
    sprintf( szbuf, "%lu", (unsigned long)numRows()-1 );
    setText( numRows()-1, COL_INDEX, szbuf );

    updateEntry( numRows() - 1, e );
}

static QColor colourFor( Symo::TagStatus st )
{
    if ( st == Symo::PLAIN )
	return QColor();

    if ( st == Symo::SUSPICIOUS )
	return Qt::red;

    // Edited
    return Qt::blue;
}

class ColouredItem: public QTableItem
{
    QColor m_colour;
public:
    ColouredItem( QTable *table, const QString& text, QColor col )
	: QTableItem( table, OnTyping, text, QPixmap() ), m_colour(col)
	{
	    setReplaceable(false);
	}
    virtual void paint( QPainter *p, const QColorGroup& cg, const QRect& cr,
			bool selected )
	{
	    QColorGroup cg2 = cg;
	    cg2.setColor( QColorGroup::Text, m_colour );
	    QTableItem::paint( p, cg2, cr, selected );
	}
    virtual int alignment() const { return AlignLeft; }
};

void SymoGrid::columnClicked( int col )
{
    if ( col == m_lastSortCol )
	m_sortAsc = !m_sortAsc;
    else
    {
	m_lastSortCol = col;
	m_sortAsc = true;
    }
    sortColumn( col, m_sortAsc, true );
    buildIndexMap();
    horizontalHeader()->setSortIndicator( col, m_sortAsc );
}

void SymoGrid::updateIndex( unsigned int index )
{
    unsigned int row;

    if ( m_map.empty() )
	row = index;
    else
	row = m_map[index];
	
    updateEntry( row, m_symo->entry(index) );
}

void SymoGrid::buildIndexMap()
{
    m_map.clear();

    for ( unsigned int i=0; i<(unsigned int)numRows(); ++i )
    {
	unsigned int index = atoi( text(i,COL_INDEX) );
	m_map[index] = i;
    }
}

/** @todo Only update if text changes */
void SymoGrid::updateEntry( unsigned int row, const Symo::Entry& e )
{
    setItem( row, COL_FILENAME, 
	     new ColouredItem( this,
			       e.filename.c_str(),
			       colourFor(e.filenameStatus) ) );
    setItem( row, COL_ARTIST, 
	     new ColouredItem( this,
			       e.artist.c_str(),
			       colourFor(e.artistStatus) ) );
    setItem( row, COL_TITLE, 
	     new ColouredItem( this,
			       e.title.c_str(),
			       colourFor(e.titleStatus) ) );
    setItem( row, COL_ALBUM, 
	     new ColouredItem( this,
			       e.album.c_str(),
			       colourFor(e.albumStatus) ) );
    setItem( row, COL_YEAR, 
	     new ColouredItem( this,
			       e.year.c_str(),
			       colourFor(e.yearStatus) ) );
    setItem( row, COL_GENRE,
	     new ColouredItem( this, e.genre.c_str(),
			       colourFor(e.genreStatus) ) );

    setItem( row, COL_TRACK,
	     new NumericTableItem( this, QTableItem::OnTyping,
				   e.trackno.c_str() ) );

    char szbuf[16];
    sprintf( szbuf, "%lu", (unsigned long)e.sz );

    setItem( row, COL_SIZE,
	     new NumericTableItem( this, QTableItem::Never, szbuf ) );

    char tmbuf[32];
    strftime( tmbuf, sizeof(tmbuf), "%H:%M:%S %d-%b-%Y", 
	      localtime( &e.mtime ) );

    setItem( row, COL_MTIME,
	     new HiddenNumericTableItem( this, QTableItem::Never, tmbuf, e.mtime ) );

    sprintf( tmbuf, "%d:%02d", e.sec/60, e.sec%60 );
    setItem( row, COL_MMSS,
	     new HiddenNumericTableItem( this, QTableItem::Never, tmbuf, e.sec ) );

    setItem( row, COL_BITRATE,
	     new QTableItem( this, QTableItem::Never, e.bitrate.c_str() ) );

    setItem( row, COL_EDITED,
	     new QTableItem( this, QTableItem::Never, "" ) );

    if ( e.edited )
	setPixmap( row, COL_EDITED, m_pmEdited );
    else
	setPixmap( row, COL_EDITED, QPixmap() );
}


QString NumericTableItem::key() const
{
    QString s = "0000000000" + text();
    return s.right(10);
}


HiddenNumericTableItem::HiddenNumericTableItem( QTable *table,
						QTableItem::EditType et,
						const QString& text,
						unsigned int value )
    : QTableItem( table, et, text ), m_value(value)
{
}

QString HiddenNumericTableItem::key() const
{
    QString s;
    s.sprintf( "%010d", m_value );
    return s;
}


void SymoGrid::checkUpdate( int row, int col, QString value, 
			    std::set<unsigned int>& needs_updating )
{
    if ( col == COL_EDITED || col == COL_SIZE || col == COL_INDEX
	|| col == COL_MTIME )
	return; // not editable

    int index = atoi( text(row,COL_INDEX) );

    Symo::Entry se = m_symo->entry(index); // structure copy
    
    switch ( col )
    {
    case COL_ARTIST:
	se.artist = value;
	break;
    case COL_ALBUM:
	se.album = value;
	break;
    case COL_TITLE:
	se.title = value;
	break;
    case COL_YEAR:
	se.year = value;
	break;
    case COL_GENRE:
	se.genre = value;
	break;
    case COL_FILENAME:
	se.filename = value;
	break;
    case COL_TRACK:
	se.trackno = value;
	break;
    }

    m_symo->changeEntry( index, se );
    needs_updating.insert( row );
}

void SymoGrid::onValueChanged( int row, int col )
{
    QString value = text( row, col );

    if ( value == m_oldText )
	return;

    std::set<unsigned int> needs_updating;

    checkUpdate( row, col, value, needs_updating );

    for ( int i=0; ; ++i )
    {
	QTableSelection qts = selection(i);

	if ( !qts.isActive() )
	    break;

	for ( int col = qts.leftCol(); col <= qts.rightCol(); ++col )
	    for ( int row = qts.topRow(); row <= qts.bottomRow(); ++row )
		checkUpdate( row, col, value, needs_updating );
    }

    // needs_updating contains the grid rows that need a redraw
    for ( std::set<unsigned int>::iterator i = needs_updating.begin();
	  i != needs_updating.end();
	  ++i )
    {
	int index = atoi( text(*i,COL_INDEX) );
	
	updateEntry( *i, m_symo->entry(index) );
    }
}
	
void SymoGrid::fitAll()
{
    for ( int i=0; i<numCols(); ++i )
    {
	if ( i != COL_INDEX )
	    adjustColumn(i);
    }
}

void SymoGrid::endEdit( int row, int col, bool accept, bool replace )
{
    if ( accept )
	m_oldText = text(row,col);

    QTable::endEdit( row, col, accept, replace );
}

void SymoGrid::doAutoFixup()
{
    std::set<unsigned int> needs_updating;

    needs_updating.insert( currentRow() );

    for ( int i=0; ; ++i )
    {
	QTableSelection qts = selection(i);

	if ( !qts.isActive() )
	    break;

	for ( int row = qts.topRow(); row <= qts.bottomRow(); ++row )
	    needs_updating.insert( row );
    }

    for ( std::set<unsigned int>::iterator i = needs_updating.begin();
	  i != needs_updating.end();
	  ++i )
    {
	int index = atoi( text(*i,COL_INDEX) );
	
	m_symo->autoFixup(index);
	updateEntry( *i, m_symo->entry(index) );
    }
}

/** Return starts in-place editing (why isn't this the default?) */
void SymoGrid::keyPressEvent( QKeyEvent *e )
{
//    printf( "key=%d state=%d\n", e->key(), e->state() );

    if ( e->key() == Key_Return || e->key() == Key_Enter )
    {
	// Fake an F2, which does start editing
	QKeyEvent e2( QKeyEvent::KeyPress, Key_F2, 0, e->state() );
	QTable::keyPressEvent(&e2);
    }
    else if ( e->key() == Key_F && e->state() == ControlButton )
    {
	// ^F = auto-fixup
	doAutoFixup();
    }
    else if ( e->key() == Key_P && e->state() == ControlButton )
    {
	// ^P = play
	int row = currentRow();
	int index = atoi( text(row,COL_INDEX) );
	m_symo->play(index);
    }
    else
	QTable::keyPressEvent(e);
}


        /* SymoWindow */


SymoWindow::SymoWindow( Symo *symo, QWidget *parent )
    : QWidget( parent ), m_symo(symo)
{
    setFocusPolicy(StrongFocus);

    m_grid = new SymoGrid( symo, this );

    m_progress = new QProgressDialog( this );

    setCaption("SYMO " VERSION);

    m_errors = new QTextView(0);
    m_errors->setCaption("SYMO errors");

    m_progress->setCaption("SYMO progress");
    m_progress->setLabelText("Scanning files...");
    symo->read(this);

    for ( unsigned int i = 0; i < symo->entries(); i++ )
	m_grid->addEntry( symo->entry(i) );

    m_progress->reset();

    m_grid->fitAll();

    QVBoxLayout *lay1 = new QVBoxLayout(this);
    QHBoxLayout *lay2 = new QHBoxLayout(3);

    QPushButton *cancel = new QPushButton("Cancel",this);
    QPushButton *ok = new QPushButton("Write changes",this);

    lay2->addStretch(1);
    lay2->addWidget(cancel);
    lay2->addWidget(ok);

    lay1->addWidget( m_grid, 1 );
    lay1->addSpacing(3);
    lay1->addLayout( lay2 );
    lay1->addSpacing(3);

    lay1->activate();

    connect( ok, SIGNAL(clicked()),
	     this, SLOT(onOK()) );
    connect( cancel, SIGNAL(clicked()),
	     this, SLOT(onCancel()) );
}

SymoWindow::~SymoWindow()
{
}

void SymoWindow::onOK()
{
    m_progress->setLabelText("Writing changes...");
    m_symo->write(this);
    m_progress->reset();
}

void SymoWindow::onCancel()
{
    if ( okToQuit() )
	qApp->quit();
}

bool SymoWindow::onActivity( unsigned int i, unsigned int n )
{
//    printf("activity %d/%d\n", i, n );
    m_progress->setTotalSteps(n);
    m_progress->setProgress(i);
    qApp->processEvents();
    return m_progress->wasCancelled();
}

void SymoWindow::onError( string s )
{
    m_errors->append( s.c_str() );
    m_errors->show();
}

void SymoWindow::onEntryChanged( unsigned int index )
{
    // Called when changes are committed to disk (i.e. when status changes
    // from edited to plain).

    m_grid->updateIndex( index );
}

bool SymoWindow::okToQuit()
{
    QString s;
    unsigned int edited = m_symo->editedCount();

    if ( edited == 0 )
	return true;

    if ( edited == 1 )
	s = "Tag changes are unsaved.";
    else
	s.sprintf( "Tag changes to %d files are unsaved.", edited );

    switch ( QMessageBox::information( this, "SYMO",
				       s + " Do you want to save before exiting?",
				       "&Save", "&Don't save", "&Cancel",
				       0,
				       2 ) )
    {
    case 0: // Save, Alt-S, Enter
	onOK();
	return true;
    case 1: // Don't save, Alt-D
	return true;
    case 2: // Cancel, Alt-C, Esc
	return false;
    }
    return false;
}

void SymoWindow::closeEvent( QCloseEvent *e )
{
    if ( okToQuit() )
	e->accept();
}

void SymoWindow::focusInEvent( QFocusEvent *e )
{
    if ( e->gotFocus() )
	m_grid->setFocus();
}

int main( int argc, char *argv[] )
{
    QApplication a( argc, argv );

    QApplication::setStyle( new QWindowsStyle() );

    Symo *symo = new Symo();

    for ( int i=1; i<argc; i++ )
	symo->addFile( argv[i] );

    SymoWindow *w = new SymoWindow(symo);

    // For want of anything better...
    w->setIcon( QPixmap( (const char**)mini_exclam_xpm ) );

    a.setMainWidget( w );

    w->show();

    return a.exec();
}

/* eof */
